// This program takes a trained model in the form of a serialized binary protocol buffer (binaryproto)
// and converts it into the equivalent HDF5 format.  The program may be invoked in one of two ways:
//
//    convertToHdf5 <model definition> <Caffe model> <HDF5 equivalent of Caffe model>
//
// or
//
//    convertToHdf5 <model definition> <layer mapping> <Caffe model> <HDF5 equivalent of Caffe model>
//
// An example of the first invocation is
//
//    convertToHdf5 VGG_ILSVRC_16_layers.prototxt VGG_ILSVRC_16_layers.caffemodel VGG_ILSVRC_16_layers.caffemodel.h5
//
// where
//    VGG_ILSVRC_16_layers.prototxt -> model definition file for VGG 16
//    VGG_ILSVRC_16_layers.caffemodel -> VGG 16 Caffe model in serialized binary protocol buffer form
//    VGG_ILSVRC_16_layers.caffemodel.h5 -> VGG 16 Caffe model in HDF5 format
//
// An example of the second invocation is
//
//    convertToHdf5 ResNet-50-deploy.prototxt ResNet-50-mapping.txt ResNet-50-model.caffemodel ResNet-50-model.caffemodel.h5
//
// where
//    ResNet-50-deploy.prototxt -> model definition file for ResNet 50
//    ResNet-50-mapping.txt -> text file allowing renaming of original ResNet 50 layers to new names.  Each line of this file
//                               should have the form
//                                  original_layer_name     new_layer_name
//                               where original_layer_name is taken from a layer defined in ResNet-50-deploy.prototxt, and
//                               new_layer_name is the correponding new layer name.
//    ResNet-50-model.caffemodel -> ResNet 50 Caffe model in serialized binary protocol buffer form
//    ResNet-50-model.caffemodel.h5 -> ResNet 50 Caffe model in HDF5 format 

 
#include <string> 
#include <vector> 
 
#include "caffe/blob.hpp" 
#include "caffe/net.hpp" 
#include "caffe/proto/caffe.pb.h" 
#include "caffe/util/io.hpp" 
#include "caffe/util/hdf5.hpp" 
#include "caffe/layer.hpp" 
#include "caffe/layers/conv_layer.hpp" 
#include "caffe/layers/lrn_layer.hpp" 
#include "caffe/layers/pooling_layer.hpp" 
#include "caffe/layers/relu_layer.hpp" 
#include "caffe/layers/sigmoid_layer.hpp" 
#include "caffe/layers/softmax_layer.hpp" 
#include "caffe/layers/tanh_layer.hpp" 
#include "caffe/layers/inner_product_layer.hpp" 
#include "caffe/layers/batch_norm_layer.hpp"
#include "caffe/layers/scale_layer.hpp"
 
using namespace caffe;  // NOLINT(build/namespaces) 

//#define DEBUG_OUTPUT
 
void extractLayerBlobs(hid_t &h5file, Layer<float> *layer, const string& layer_name); 
 
int main(int argc, char** argv) 
{ 
   caffe::GlobalInit(&argc, &argv); 
   Caffe::set_mode(Caffe::CPU); 
 
   // parse command line arguments
   const char *network_params = NULL;
   const char *network_snapshot = NULL;
   const char *hdf5_output_fn = NULL;
   const char *layer_mapping = NULL;
   switch (argc) 
   {
      case 4:
      {
         network_params = argv[1];
         network_snapshot = argv[2];
         hdf5_output_fn = argv[3];
         break;
      }
      case 5:
      {
         network_params = argv[1];
         layer_mapping = argv[2];
         network_snapshot = argv[3];
         hdf5_output_fn = argv[4];
         break;
      }
      default:
      {
         std::cout << "INFO: Acceptable usage:" << std::endl; 
         std::cout << "    " << argv[0] << " model_definition.prototxt model_parameters.caffemodel model_parameters.caffemodel.h5" << std::endl;
         std::cout << " or " << std::endl;
         std::cout << "    " << argv[0] << " model_definition.prototxt layer-mapping.txt model_parameters.caffemodel model_parameters.caffemodel.h5" << std::endl;
         exit(1); 
      }
   }

   // read layer name mapping file (if given)
   vector<string> mapLayerName;
   mapLayerName.clear();
   vector<string> originalLayerName;
   originalLayerName.clear();
   if (layer_mapping) 
   {
      std::string line;
      std::ifstream infile(layer_mapping);

      // fail if file does not exist
      if (!infile.is_open()) 
      {
         std::cout << "ERROR: Unable to open/read file " << layer_mapping << std::cout;
         exit(1);
      }
      
      while (std::getline(infile, line))
      {
         std::istringstream iss(line);
         string str1, str2;
         if (!(iss >> str1  >> str2)) 
         { 
            std::cout << "ERROR: layer mapping file must have two entries per line" << std::endl;
            exit(1);
            break; 
         }
         originalLayerName.push_back(str1);
         mapLayerName.push_back(str2);    
      }
   }
   
   // create output file
   hid_t file_id = H5Fcreate(hdf5_output_fn, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT); 
 
   // read network definition and trained parameters
   shared_ptr<Net<float> > caffe_net; 
   caffe_net.reset(new Net<float>(network_params, caffe::TEST)); 
   caffe_net->CopyTrainedLayersFrom(network_snapshot); 

   // create HDF5 structure to match the output of Caffe if using "snapshot_format: HDF5" in the solver definition file
   hid_t data_hid = H5Gcreate2(file_id, "data", H5P_DEFAULT, H5P_DEFAULT,H5P_DEFAULT);
 
   // get layer names
   const vector<shared_ptr<Layer<float> > >& layers = caffe_net->layers(); 
   const vector<string> & layer_names = caffe_net->layer_names(); 
   
   // loop over layers and extract the relevant blobs
   for (int ii = 0; ii < layer_names.size(); ii++) 
   { 
      hid_t layer_data_hid = -1;
      // see if layer is renamed
      for (int jj = 0; jj < originalLayerName.size(); jj++) 
      {
         if (originalLayerName[jj].compare(layer_names[ii]) == 0) 
         {
            layer_data_hid =  H5Gcreate2(data_hid, mapLayerName[jj].c_str(),H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
#ifdef DEBUG_OUTPUT
            std::cout << "************* Layer " << layer_names[ii] << " mapped to " << mapLayerName[jj] << " in HDF5 file ***************" << std::endl;
#endif // DEBUG_OUTPUT
            break;
         }
      }
      
      // use original layer name if renaming not required
      if (layer_data_hid < 0)
         layer_data_hid = H5Gcreate2(data_hid, layer_names[ii].c_str(),H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);

      // extract blobs if
      //    - layer has associated blobs
      //    - layer is supported by TKDL
      // otherwise, print failure message
      if (InnerProductLayer<float> *layer = dynamic_cast<InnerProductLayer<float> *>(layers[ii].get())) 
      { 
         std::cout << "Extracting blob(s) from InnerProduct layer " << layer_names[ii] << std::endl; 
         extractLayerBlobs(layer_data_hid, layer, layer_names[ii]); 
      } 
      else 
      {
         if (ConvolutionLayer<float> *layer = dynamic_cast<ConvolutionLayer<float> *>(layers[ii].get())) 
         { 
            std::cout << "Extracting blob(s) from Convolution layer " << layer_names[ii] << std::endl; 
            extractLayerBlobs(layer_data_hid, layer, layer_names[ii]); 
         } 
         else 
         {
            if (BatchNormLayer<float> *layer = dynamic_cast<BatchNormLayer<float> *>(layers[ii].get())) 
            {
               std::cout << "Extracting blob(s) from BatchNorm layer " << layer_names[ii] << std::endl; 
               extractLayerBlobs(layer_data_hid, layer, layer_names[ii]); 
            } 
            else 
            {
               if (ScaleLayer<float> *layer = dynamic_cast<ScaleLayer<float> *>(layers[ii].get())) 
               {
                  std::cout << "Extracting blob(s) from Scale layer " << layer_names[ii] << std::endl; 
                  extractLayerBlobs(layer_data_hid, layer, layer_names[ii]); 
               } 
               else 
               { 
                  std::cout << "No parameters or TKDL does not (yet) support operations for layer " << layer_names[ii] << std::endl; 
               }
            }
         }
      }
      H5Gclose(layer_data_hid);
   }  
 
  H5Gclose(data_hid);
  H5Fclose(file_id); 
  return 0; 
} 
  
void extractLayerBlobs(hid_t &h5file, Layer<float> *layer, const string& layer_name) 
{ 
   vector<shared_ptr<Blob<float> > >& blobs = layer->blobs(); 
   if (blobs.size() == 0) 
   { 
      std::cout << "Layer " << layer_name << " has no blobs!!!" << std::endl; 
      return; 
   }  

   // loop over layer blobs and write to HDF5 file
   for (int ii = 0; ii < blobs.size(); ii++) 
   {
      string linkName = std::to_string(ii);
      std::cout << layer_name << ": blob # " << linkName << std::endl;
      hdf5_save_nd_dataset(h5file, linkName.c_str(), *blobs[ii]); 
#ifdef DEBUG_OUTPUT
      const float *dataPtr = (*blobs[ii]).cpu_data();
      std::cout << std::endl;
      for (int jj = 0; jj < 5; jj++)
         std::cout << dataPtr[jj] << " ";
      std::cout << std::endl;
#endif
   }
}
